#include <cmath>
#include <fstream>
#include <initializer_list>
#include <iostream>
#include <list>
#include <type_traits>
#include "gfx.h"
#include "libsdl.h"
#include "events.h"
#include "main.h"
#include "prog.h"


#if MANIFEST_WINDOW_INITIAL_STATE == 1 && MANIFEST_WINDOW_RESIZABLE == 0
#error Fixed size window can`t be maximized. (MANIFEST_WINDOW_INITIAL_STATE == 1 && MANIFEST_WINDOW_RESIZABLE == 0)
#endif


namespace GFX
{
	std::list<Shader::pack> Shader::load_list;
	static SDL_Window *winhandle;
	static SDL_GLContext conhandle;

    SDL_Window *WinHandle()
    {
    	return winhandle;
    }

    SDL_GLContext ConHandle()
    {
    	return conhandle;
    }

	Shader::Shader(const char *vsrc, const char *fsrc)
	  : prog(0), vsh(0), fsh(0)
	{
		load_list.push_back({this, vsrc, fsrc});
	}

    void Shader::LoadShaderList()
	{
		static unsigned int id = 0;

		for (auto &it : load_list)
		{
			it.p->prog = glCreateProgram();
			it.p->vsh = glCreateShader(GL_VERTEX_SHADER);
			it.p->fsh = glCreateShader(GL_FRAGMENT_SHADER);

			if (!it.p->prog)
				Main::Exit("glCreateProgram() failed.");
			if (!(it.p->vsh && it.p->fsh))
				Main::Exit("glCreateShader() failed.");

			glAttachShader(it.p->prog, it.p->vsh);
			glAttachShader(it.p->prog, it.p->fsh);
			glShaderSource(it.p->vsh, 1, &it.vsrc, 0);
			glShaderSource(it.p->fsh, 1, &it.fsrc, 0);
			glCompileShader(it.p->vsh);
			glCompileShader(it.p->fsh);

			GLint vstat, fstat;
			glGetShaderiv(it.p->vsh, GL_COMPILE_STATUS, &vstat);
			glGetShaderiv(it.p->fsh, GL_COMPILE_STATUS, &fstat);

			if (vstat != GL_TRUE || fstat != GL_TRUE)
			{
				char *vlog, *flog;
				glGetShaderiv(it.p->vsh, GL_INFO_LOG_LENGTH, &vstat);
				glGetShaderiv(it.p->fsh, GL_INFO_LOG_LENGTH, &fstat);
				if (vstat)
				{
					vlog = new char[vstat];
					glGetShaderInfoLog(it.p->vsh, vstat, 0, vlog);
				}
				else
					vlog = (char *) "<empty>";
				if (fstat)
				{
					flog = new char[fstat];
					glGetShaderInfoLog(it.p->fsh, fstat, 0, flog);
				}
				else
					flog = (char *) "<empty>";

				Main::Exit(Str({"Shader #", id, " compilation failed:",
								(vstat != GL_TRUE) ? "\nVertex shader: NOT OK":"\nVertex shader: OK",
								(fstat != GL_TRUE) ? "\nFragment shader: NOT OK":"\nFragment shader: OK",
								"\n\n  Vertex log:\n", vlog, "\n  Fragment log:\n", flog}));
			}
			glLinkProgram(it.p->prog);
			glGetProgramiv(it.p->prog, GL_LINK_STATUS, &vstat);
			if (vstat != GL_TRUE)
			{
				glGetProgramiv(it.p->prog, GL_INFO_LOG_LENGTH, &vstat);
				char *log;
				if (vstat)
				{
					log = new char[vstat];
					glGetProgramInfoLog(it.p->prog, vstat, 0, log);
				}
				else
					log = (char *) "<empty>";

				Main::Exit(Str({"Shader #", id, " linking failed.\n\n  Log:\n",log}));
			}

			id++;
		}

		load_list.clear();
	}

	const float (*const CameraController::r)[3] = Quaternion::GetMassTransformMatrix3x3();

	GLuint BindingContainer::vao_binding = 0;
	GLuint BindingContainer::vbo_binding = 0;

	ImageData::ImageData(const char *fname)
	{
		std::ifstream inp_file(fname, std::ios_base::in | std::ios_base::binary);
		if (!inp_file)
			Main::Exit(Str({"Can't load ", fname, ": File not found."}));
		inp_file >> std::noskipws;
		unsigned char idfieldlen, tmp;
		bool swap_x, swap_y;

		inp_file >> idfieldlen;																							// 00
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Str({"Can't load ", fname, ": Files with colormaps are not supported."})); // 01
		inp_file >> tmp; if (tmp != 0x02) Main::Exit(Str({"Can't load ", fname, ": File must uncompressed 32 bpp TrueColor image (type check failure)."})); // 02
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Str({"Can't load ", fname, ": Files with colormaps are not supported (1)."})); // 03
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Str({"Can't load ", fname, ": Files with colormaps are not supported (2)."})); // 04
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Str({"Can't load ", fname, ": Files with colormaps are not supported (3)."})); // 05
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Str({"Can't load ", fname, ": Files with colormaps are not supported (4)."})); // 06
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Str({"Can't load ", fname, ": Files with colormaps are not supported (5)."})); // 07
		inp_file.ignore(4);                                                                                             // 08
																														// 09
																														// 0a
																														// 0b
		inp_file >> tmp; img_w = tmp;																					// 0c
		inp_file >> tmp; img_w |= tmp << 8;																				// 0d
		inp_file >> tmp; img_h = tmp;																					// 0e
		inp_file >> tmp; img_h |= tmp << 8;																				// 0f
		inp_file >> tmp; if (tmp != 0x20) Main::Exit(Str({"Can't load ", fname, ": File must uncompressed 32 bpp TrueColor image (bpp check failure)."})); // 10
		inp_file >> tmp; if ((tmp&15)!=8) Main::Exit(Str({"Can't load ", fname, ": File must uncompressed 32 bpp TrueColor image (alpha check failure)."})); // 11

		swap_x = tmp & 0x10; swap_y = !(tmp & 0x20);

		inp_file.ignore(idfieldlen);

		if (!inp_file) Main::Exit(Str({"Can't load ", fname, ": Unexpected end of the file (header)."}));

		data = new unsigned char[img_w * img_h][4];
		int x, y, xend, yend, xstep, ystep, xbegin;

		if (!swap_x)
		{
			xbegin = 0;
			xend = img_w;
			xstep = 1;
		}
		else
		{
			xbegin = img_w - 1;
			xend = -1;
			xstep = -1;
		}
		if (!swap_y)
		{
			y = 0;
			yend = img_h;
			ystep = 1;
		}
		else
		{
			y = img_h - 1;
			yend = -1;
			ystep = -1;
		}


		while (y != yend)
		{
			x = xbegin;
			while (x != xend)
			{
				auto &ref = data[x + img_w * y];
				inp_file >> ref[2];
				inp_file >> ref[1];
				inp_file >> ref[0];
				inp_file >> ref[3];
				x += xstep;
			}
			y += ystep;
		}
		if (!inp_file) Main::Exit(Str({"Can't load ", fname, ": Unexpected end of the file."}));

		bytelen = img_w * img_h * 4 * sizeof (unsigned char);
	}

	float Quaternion::t[3][3];

	static int win_w, win_h, win_w_half, win_h_half, win_w_mo2, win_h_mo2;
	float win_w_div_h;

	int WindowW()       {return win_w;}
	int WindowH()       {return win_h;}
	float WindowRatio() {return win_w_div_h;}
	int WindowW_half()  {return win_w_half;}
	int WindowH_half()  {return win_h_half;}
	int WindowW_m_of_2(){return win_w_mo2;}
	int WindowH_m_of_2(){return win_h_mo2;}

	void CheckWindowSize()
	{
		int w, h;
		SDL_GetWindowSize(winhandle, &w, &h);
		if (w == win_w && h == win_h)
			return;
		win_w = w;
		win_h = h;
		win_w_half = (win_w + 1) / 2;
		win_h_half = (win_h + 1) / 2;
		win_w_mo2 = win_w_half * 2;
		win_h_mo2 = win_h_half * 2;
		win_w_div_h = float(w) / float(h);
		Prog::Resize(w, h);
	}

	void Init(int w, int h)
	{
		EXEC_ONCE;

		static constexpr uint32_t sdlwin_flags = SDL_WINDOW_OPENGL
											   | (MANIFEST_WINDOW_RESIZABLE ? SDL_WINDOW_RESIZABLE : 0)
											   | SDL_WINDOW_FULLSCREEN;

		static char openglfailstring[] = "OpenGL ?.? (compatibility profile) context init failed.\n"
										 "\n\n\n"
										 "1. Update your videocard driver.\n"
										 "2. Force your driver to use best available videocard for this application.\n"
										 "If it still complains about your OpenGL, then your videocard\n"
										 "is probably too old for this program.\n";

		int msample = MANIFEST_MSAA;
		if (msample & (msample - 1))
			Main::Exit("Requested MSAA is not a power of two.");

		openglfailstring[7] = MANIFEST_OPENGL_MAJOR + '0';
		openglfailstring[9] = MANIFEST_OPENGL_MINOR + '0';


		glewExperimental = 1;

		SDL_SetHint(SDL_HINT_RENDER_VSYNC, "1");
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, MANIFEST_OPENGL_MAJOR);
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, MANIFEST_OPENGL_MINOR);
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, MANIFEST_OPENGL_CORE_CONTEXT ? SDL_GL_CONTEXT_PROFILE_CORE : SDL_GL_CONTEXT_PROFILE_COMPATIBILITY);
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, MANIFEST_OPENGL_FORWARD_COMPAT ? SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG : 0);

		if (msample > 1)
		{
			SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
			SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, msample);
		}

		SDL_DisplayMode mode;
		SDL_GetDesktopDisplayMode(0, &mode);
	  cont_creation:

		if (!(winhandle = SDL_CreateWindow(MANIFEST_APPNAME, (mode.w - MANIFEST_WINDOW_W) / 2, (mode.h - MANIFEST_WINDOW_H) / 2, mode.w, mode.h, sdlwin_flags)))
			Main::Exit(Str({"Window creation failed: `", SDL_GetError(), '`'}));

		SDL_SetWindowMinimumSize(winhandle, MANIFEST_WINDOW_W, MANIFEST_WINDOW_H);

		if (!(conhandle = SDL_GL_CreateContext(winhandle)))
		{
			if (msample > 1)
			{
				SDL_DestroyWindow(winhandle);
				msample /= 2;
				if (msample > 1)
				{
					SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, msample);
				}
				else
					SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 0);
				goto cont_creation;
			}
			Main::Exit(openglfailstring);
		}

		SDL_GL_SetSwapInterval(1);

		if (glewInit() != GLEW_OK)
			Main::Exit("GLEW init failed.");


		win_w = w;
		win_h = h;
		win_w_half = (win_w + 1) / 2;
		win_h_half = (win_h + 1) / 2;
		win_w_mo2 = win_w_half * 2;
		win_h_mo2 = win_h_half * 2;
		win_w_div_h = float(w) / float(h);

		glEnable(GL_BLEND);
		glDepthFunc(GL_GEQUAL);
		glClearDepth(-1);

		glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA);
	}
}