#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <csignal>
#include <exception>
#include "libglew.h"
  #define AUDIO_H_INIT_ACCESS
#include "audio.h"
  #define EVENTS_H_INIT_ACCESS
#include "events.h"
#include "prog.h"
#include "main.h"
  #define GFX_H_INIT_ACCESS
#include "gfx.h"
#include "utils.h"

namespace Main
{
	void Msg(const char *txt)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_INFORMATION, MANIFEST_APPNAME, txt, 0);
	}

	[[noreturn]] void Exit()
	{
		std::exit(0);
	}

	[[noreturn]] void Exit(const char *txt)
	{
		SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, MANIFEST_APPNAME " - Error", txt, 0);
		std::exit(0);
	}

	static int r_argc;
	static const char *const *r_argv;

	int argc()
	{
		return r_argc;
	}
	const char *const *argv()
	{
		return r_argv;
	}

	Main::StateFunc::PtrType current_state = Prog::States::Boot;

	static void Init()
	{
		EXEC_ONCE;

		static constexpr uint32_t sdlinit_flags = SDL_INIT_VIDEO | SDL_INIT_EVENTS | SDL_INIT_NOPARACHUTE; // Because parachutes are for noobs
		if (SDL_Init(sdlinit_flags))
			Main::Exit(Str({"SDL init failed: `", SDL_GetError(), '`'}));
		std::atexit(SDL_Quit);

		Events::Init();
		GFX::Init(MANIFEST_WINDOW_W, MANIFEST_WINDOW_H);
		Audio::Init();
		Prog::Init(MANIFEST_WINDOW_W, MANIFEST_WINDOW_H);
	}
}

int main(int argc, char **argv)
{
	EXEC_ONCE;

	auto SignalHandler = [](int sig) -> void
	{
		switch (sig)
		{
		  case SIGINT: case SIGTERM:
		  #ifdef SIGBREAK
		  case SIGBREAK:
		  #endif
			Main::Exit();
			break;

		  case SIGILL:
			Main::Exit("SIGILL: Illegal instruction.");
			break;

		  case SIGFPE:
			Main::Exit("SIGFPE: Floating point error.");
			break;

		  case SIGSEGV:
			Main::Exit("SIGSEGV: Segmentation violation.");
			break;

		  case SIGABRT:
			Main::Exit("SIGABRT: Abnormal termination.");
			break;

		  default:
			Main::Exit("Unknown signal.");
			break;
		}
	};

	std::signal(SIGINT  , SignalHandler);
	std::signal(SIGTERM , SignalHandler);
	std::signal(SIGILL  , SignalHandler);
	std::signal(SIGFPE  , SignalHandler);
	std::signal(SIGSEGV , SignalHandler);
	std::signal(SIGABRT , SignalHandler);
	#ifdef SIGBREAK
	std::signal(SIGBREAK, SignalHandler);
	#endif

	std::set_terminate(
	[]{
		try {throw;}
		catch (const std::exception &e) {Main::Exit(Str({"Standard exception: `", e.what(), "`."}));}
		catch (const char *p) {Main::Exit(Str({"Exception: `", p, "`."}));}
		catch (...) {Main::Exit(Str({"Unknown exception."}));}
	});
	std::set_unexpected(
	[]{
		try {throw;}
		catch (const std::exception &e) {Main::Exit(Str({"Unexpected standard exception: `", e.what(), "`."}));}
		catch (const char *p) {Main::Exit(Str({"Unexpected exception: `", p, "`."}));}
		catch (...) {Main::Exit(Str({"Unexpected unknown exception."}));}
	});

	Main::r_argc = argc;
	Main::r_argv = argv;

	Main::Init();

	while (Main::current_state) Main::current_state = (Main::StateFunc::PtrType)Main::current_state();

	Main::Exit();

	return 0; // Macro replaced to be `SDL_main` there is no automatic return 0 here!
}