#include <cstring>
#include <fstream>
#include <initializer_list>
#include <list>
#include "audio.h"
#include "gfx.h"
#include "main.h"
#include "prog.h"
#include "shaders.h"
#include "utils.h"

class Box
{
  public:
	int type;
	int x, y;
	float xv, yv;
	int charge;
	Box(int xpos, int ypos, int type_id)
	{
		type = type_id;
		xv = yv = 0;
		charge = 0;
		switch (type)
		{
		  case 0:
		  	x = xpos * 48 + 24;
			y = ypos * 48 + 24;
			break;
		  case 1:
			x = xpos * 48 + 24;
			y = ypos * 48 + 48;
			break;
		  case 2:
			x = xpos * 48 + 48;
			y = ypos * 48 + 48;
			break;
		  case 3:
			x = xpos * 48 + 24*3;
			y = ypos * 48 + 24;
			break;
		  default:
			Main::Exit("Incorrect box type.");
			break;
		}
	}
};
class Enemy
{
  public:
  	int x, y;
  	int yv;
	bool l_or_r;
	Enemy(int xpos, int ypos, bool right)
	{
		x = xpos * 48 + 24;
		y = ypos * 48 + 24;
		yv = 0;
		l_or_r = right;
	}
};
static constexpr float block_mass[4] {1, 2, 4, 4};

namespace Maps
{
	int id_table[256], merge_table[256];
	bool solid_table[256]{};

	class Type
	{
		const char **ptr;
	  public:
		const int w, h;
		const GFX::Vec<int, 2> spawn;
		const bool gun, gun_placed;
		const GFX::Vec<int, 2> gunpos;
		const std::list<Box> boxlist;
		const std::list<Enemy> enemylist;
		Type(std::initializer_list<const char *> data, const GFX::Vec<int, 2> &spawnp, bool have_gun, bool gun_is_placed, const GFX::Vec<int, 2> &gun_pos, std::initializer_list<Box> boxes, std::initializer_list<Enemy> enemies)
		  : w(std::strlen(*data.begin())), h(data.size()), spawn(spawnp), gun(have_gun), gun_placed(gun_is_placed), gunpos(gun_pos), boxlist(boxes.begin(), boxes.end()), enemylist(enemies.begin(), enemies.end())
		{
			ptr = new const char *[h];
			int pos = 0;
			for (auto it = data.begin(); it != data.end(); it++)
			{
				ptr[pos++] = *it;
			}
		}
		Type(const Type &) = delete;
		Type(Type &&) = delete;
		Type& operator=(const Type &) = delete;
		Type& operator=(Type &&) = delete;

		int operator()(int x, int y, char ret = 'W') const
		{
			if (x < 0 || y < 0 || x >= w || y >= h)
				return ret;
			return (unsigned char) ptr[y][x];
		}

		int AtPoint(int x, int y) const
		{
			x /= 48;
			y /= 48;
			if (x < 0 || y < 0 || x >= w || y >= h)
				return 'W';
			return (unsigned char) ptr[y][x];
		}

	};

	Type rooms[]
	{
		#include "rooms.map"
	};
}

namespace Controls
{
	static constexpr int keys = 6;
	unsigned int up      = Events::Key<'W'>(),
				 down    = Events::Key<'S'>(),
				 left    = Events::Key<'A'>(),
				 right   = Events::Key<'D'>(),
				 use     = Events::Key<'E'>(),
				 pause   = Events::Key_Esc ();

	const char *names[keys];
	void UpdateNames()
	{
		names[0] = SDL_GetScancodeName((SDL_Scancode) up);
		names[1] = SDL_GetScancodeName((SDL_Scancode) down);
		names[2] = SDL_GetScancodeName((SDL_Scancode) left);
		names[3] = SDL_GetScancodeName((SDL_Scancode) right);
		names[4] = SDL_GetScancodeName((SDL_Scancode) use);
		names[5] = SDL_GetScancodeName((SDL_Scancode) pause);

		for (int i = 0; i < keys; i++)
		{
			if (*names[i] == '\0' || names[i] == 0)
				names[i] = "<unknown>";
		}
	}
}

namespace UniLoc
{
	int c_mf;
	int t_mf;
	int t_t;
	int t_co;
	int t_cm;
}

namespace Shaders
{
	GFX::Shader *s_c, *s_t;
}

namespace Layouts
{
	using pos3f_color4f = GFX::AttribLayout<GFX::Vec<float, 3>, GFX::Vec<float, 4>>;
	using pos3f_tex2f_co4f_cm4f = GFX::AttribLayout<GFX::Vec<float, 3>, GFX::Vec<float, 2>, GFX::Vec<float, 4>, GFX::Vec<float, 4>>;
}

namespace Sounds
{
	namespace Buffers
	{
		Audio::Buffer *startbeep;
		Audio::Buffer *door;
		Audio::Buffer *shot;
		Audio::Buffer *blhit;
		Audio::Buffer *wallhit;
		Audio::Buffer *jump;
		Audio::Buffer *kill;
		Audio::Buffer *death;
		Audio::Buffer *pickup;
		Audio::Buffer *victory;
		Audio::Buffer *maintheme;
	}
	Audio::Source *startbeep;
	Audio::Source *door;
	Audio::Source *shot1;
	Audio::Source *shot1a;
	Audio::Source *shot2;
	Audio::Source *shot2a;
	Audio::Source *blhit1;
	Audio::Source *blhit1a;
	Audio::Source *blhit2;
	Audio::Source *blhit2a;
	Audio::Source *wallhit;
	Audio::Source *wallhita;
	Audio::Source *jump;
	Audio::Source *kill;
	Audio::Source *death;
	Audio::Source *pickup;
	Audio::Source *victory;
	Audio::Source *maintheme;
}

GFX::CameraController camc;
GFX::RenderingQueue<Layouts::pos3f_color4f> *rq_c;
GFX::RenderingQueue<Layouts::pos3f_tex2f_co4f_cm4f> *rq_t;

void ScreenColor(const GFX::Vec<float, 4> &coff, const GFX::Vec<float, 4> &cmult)
{
	glUniform4f(UniLoc::t_co, coff.x, coff.y, coff.z, coff.w);
	glUniform4f(UniLoc::t_cm, cmult.x, cmult.y, cmult.z, cmult.w);
}


bool MouseInRect(const GFX::Vec<int, 2> &pos, GFX::Vec<int, 2> size)
{
	size = pos + size;
	int mx = Events::MouseX() - GFX::WindowW_half();
	int my = Events::MouseY() - GFX::WindowH_half();
	return mx >= pos.x && my >= pos.y && mx < size.x && my < size.y;
}


namespace Draw
{
	void PushC(const GFX::Vec<float, 2> &pos, GFX::Vec<float, 2> size, const GFX::Vec<float, 4> &color, float z = 0)
	{
		size = pos + size;
		rq_c->Push4({{pos.x,  pos.y,  z}, {color.x, color.y, color.z, color.w}},
					{{size.x, pos.y,  z}, {color.x, color.y, color.z, color.w}},
					{{size.x, size.y, z}, {color.x, color.y, color.z, color.w}},
					{{pos.x,  size.y, z}, {color.x, color.y, color.z, color.w}});
	}
	void PushT(const GFX::Vec<float, 2> &pos, GFX::Vec<float, 2> size, const GFX::Vec<float, 2> &tex, GFX::Vec<float, 2> tsize, float z = 0)
	{
		size = pos + size;
		tsize = tex + tsize;
		rq_t->Push4({{pos.x,  pos.y,  z}, {tex.x,   tex.y  }, {0,0,0,0}, {1,1,1,1}},
					{{size.x, pos.y,  z}, {tsize.x, tex.y  }, {0,0,0,0}, {1,1,1,1}},
					{{size.x, size.y, z}, {tsize.x, tsize.y}, {0,0,0,0}, {1,1,1,1}},
					{{pos.x,  size.y, z}, {tex.x,   tsize.y}, {0,0,0,0}, {1,1,1,1}});
	}
	void PushT(const GFX::Vec<float, 2> &pos, GFX::Vec<float, 2> size, const GFX::Vec<float, 2> &tex, GFX::Vec<float, 2> tsize, GFX::Vec<float, 2> center, float angle, float z = 0)
	{
		float xx = std::cos(angle), xy = std::sin(angle), yx = -xy, yy = xx;

		center = -center;
		size = size + center;
		tsize = tex + tsize;

		GFX::Vec<float, 2> a{center.x, center.y}, b{size.x, center.y}, c{size.x, size.y}, d{center.x, size.y};
		rq_t->Push4({{(a.x*xx+a.y*yx)+pos.x, (a.x*xy+a.y*yy)+pos.y, z}, {tex.x,   tex.y  }, {0,0,0,0}, {1,1,1,1}},
					{{(b.x*xx+b.y*yx)+pos.x, (b.x*xy+b.y*yy)+pos.y, z}, {tsize.x, tex.y  }, {0,0,0,0}, {1,1,1,1}},
					{{(c.x*xx+c.y*yx)+pos.x, (c.x*xy+c.y*yy)+pos.y, z}, {tsize.x, tsize.y}, {0,0,0,0}, {1,1,1,1}},
					{{(d.x*xx+d.y*yx)+pos.x, (d.x*xy+d.y*yy)+pos.y, z}, {tex.x,   tsize.y}, {0,0,0,0}, {1,1,1,1}});
	}
	void PushT(const GFX::Vec<float, 2> &pos, GFX::Vec<float, 2> size, const GFX::Vec<float, 2> &tex, GFX::Vec<float, 2> tsize, const GFX::Vec<float, 4> &co, const GFX::Vec<float, 4> &cm, float z = 0)
	{
		size = pos + size;
		tsize = tex + tsize;
		rq_t->Push4({{pos.x,  pos.y,  z}, {tex.x,   tex.y  }, co, cm},
					{{size.x, pos.y,  z}, {tsize.x, tex.y  }, co, cm},
					{{size.x, size.y, z}, {tsize.x, tsize.y}, co, cm},
					{{pos.x,  size.y, z}, {tex.x,   tsize.y}, co, cm});
	}
	void Text(GFX::Vec<int, 2> pos, const char *txt, const GFX::Vec<float, 4> &color, int align = -1, float z = -1)
	{
		int xp = pos.x;
		switch (align)
		{
		  case -1:
			{
				while (*txt)
				{
					switch (*txt)
					{
					  default:
						PushT({float(xp), float(pos.y)}, {24, 32}, {float(*(unsigned char *) txt % 16 * 12), float(*(unsigned char *) txt / 16 * 16)}, {12, 16}, {0,0,0,0}, color, z);
						xp += 24;
						txt++;
						break;
					  case '\n':
						pos.y += 32;
						xp = pos.x;
						txt++;
						break;
					}
				}
			}
			break;
		  case 0:
			{
				int xpp;
				const char *tmp = txt;
				int len = 0;
				while (*tmp != '\n' && *tmp != '\0')
				{
					if (*tmp >= 32)
						len++;
					tmp++;
				}
				xpp = xp - 12 * len;
				while (*txt)
				{
					switch (*txt)
					{
					  default:
						PushT({float(xpp), float(pos.y)}, {24, 32}, {float(*(unsigned char *) txt % 16 * 12), float(*(unsigned char *) txt / 16 * 16)}, {12, 16}, {0,0,0,0}, color, z);
						xpp += 24;
						txt++;
						break;
					  case '\n':
						pos.y += 32;
						txt++;
						len = 0;
						tmp = txt;
						while (*tmp != '\n' && *tmp != '\0')
						{
							if (*tmp >= 32)
								len++;
							tmp++;
						}
						xpp = xp - 12 * len;
						break;
					}
				}
			}
			break;
		  case 1:
			{
				int xpp;
				const char *tmp = txt;
				int len = 0;
				while (*tmp != '\n' && *tmp != '\0')
				{
					if (*tmp >= 32)
						len++;
					tmp++;
				}
				xpp = xp - 24 * len;
				while (*txt)
				{
					switch (*txt)
					{
					  default:
					  	PushT({float(xpp), float(pos.y)}, {24, 32}, {float(*(unsigned char *) txt % 16 * 12), float(*(unsigned char *) txt / 16 * 16)}, {12, 16}, {0,0,0,0}, color, z);
						xpp += 24;
						txt++;
						break;
					  case '\n':
						pos.y += 32;
						txt++;
						len = 0;
						tmp = txt;
						while (*tmp != '\n' && *tmp != '\0')
						{
							if (*tmp >= 32)
								len++;
							tmp++;
						}
						xpp = xp - 24 * len;
						break;
					}
				}
			}
			break;
		  default:
			Main::Exit("Invalid text alignment value.");
			break;
		}
	}
	void Mouse()
	{
		PushT({float((Events::MouseX() - GFX::WindowW_half() - 8) / 2 * 2), float((Events::MouseY() - GFX::WindowH_half() - 8) / 2 * 2)}, {48, 48}, {192, 0}, {24, 24}, {0,0,0,0}, {10,10,10,1}, -60);
	}
	void Map(int xpos, int ypos, const Maps::Type &map, char def = 'W')
	{
		int xbegin, ybegin, xend, yend;
		xbegin = (xpos - GFX::WindowW_half()) / 48 - 1;
		xend = xbegin + GFX::WindowW_m_of_2() / 48 + 3;
		ybegin = (ypos - GFX::WindowH_half()) / 48 - 1;
		yend = ybegin + GFX::WindowH_m_of_2() / 48 + 3;
		xpos = -xpos;
		ypos = -ypos;
		for (int y = ybegin; y != yend; y++)
		{
			for (int x = xbegin; x != xend; x++)
			{
				int id = map(x, y, def);
				int mid = Maps::merge_table[id];
				if (mid)
				{
					int tmp;
					tmp = Maps::merge_table[map(x, y-1)];
					bool u =  mid != tmp && (tmp >= 0 || tmp != -mid);
					tmp = Maps::merge_table[map(x, y+1)];
					bool d =  mid != tmp && (tmp >= 0 || tmp != -mid);
					tmp = Maps::merge_table[map(x-1, y)];
					bool l =  mid != tmp && (tmp >= 0 || tmp != -mid);
					tmp = Maps::merge_table[map(x+1, y)];
					bool r =  mid != tmp && (tmp >= 0 || tmp != -mid);
					tmp = Maps::merge_table[map(x-1, y-1)];
					bool ul = mid != tmp && (tmp >= 0 || tmp != -mid);
					tmp = Maps::merge_table[map(x-1, y+1)];
					bool dl = mid != tmp && (tmp >= 0 || tmp != -mid);
					tmp = Maps::merge_table[map(x+1, y-1)];
					bool ur = mid != tmp && (tmp >= 0 || tmp != -mid);
					tmp = Maps::merge_table[map(x+1, y+1)];
					bool dr = mid != tmp && (tmp >= 0 || tmp != -mid);
					int iul = u ? (l ? 4 : 1) : (l ? 2 : (ul ? 3 : 0));
					int iur = u ? (r ? 4 : 1) : (r ? 2 : (ur ? 3 : 0));
					int idl = d ? (l ? 4 : 1) : (l ? 2 : (dl ? 3 : 0));
					int idr = d ? (r ? 4 : 1) : (r ? 2 : (dr ? 3 : 0));
					PushT({float(xpos + x * 48 +  0), float(ypos + y * 48 +  0)}, {24, 24}, {float((Maps::id_table[id] - 1) * 24 +  0), float(256 +  0 + iul * 24)}, {12, 12}, -0.5);
					PushT({float(xpos + x * 48 + 24), float(ypos + y * 48 +  0)}, {24, 24}, {float((Maps::id_table[id] - 1) * 24 + 12), float(256 +  0 + iur * 24)}, {12, 12}, -0.5);
					PushT({float(xpos + x * 48 +  0), float(ypos + y * 48 + 24)}, {24, 24}, {float((Maps::id_table[id] - 1) * 24 +  0), float(256 + 12 + idl * 24)}, {12, 12}, -0.5);
					PushT({float(xpos + x * 48 + 24), float(ypos + y * 48 + 24)}, {24, 24}, {float((Maps::id_table[id] - 1) * 24 + 12), float(256 + 12 + idr * 24)}, {12, 12}, -0.5);
				}
				else
				{
					PushT({float(xpos + x * 48), float(ypos + y * 48)}, {24, 24}, {float((Maps::id_table[id] - 1) * 24), 256}, {12, 12}, -0.5);
				}
			}
		}
	}
}

static constexpr int menu_w = 256, menu_h = 48, menu_y = 64;
static constexpr int pl_y = 128;

namespace Player
{
	int x, y;
	int xv, yv;
	Maps::Type *map;
	bool l_or_r;
	int anim_state;
	bool nextmap;
	bool gun, gun_placed, gunpickuptext;
	GFX::Vec<int, 2> placedgunpos;
	float gunrot;

	bool pr[2];
	GFX::Vec<float, 2> prpos[2];
	float pra[2];

	bool paused, nopause;
	float hlight[4];
	int menu_choice;
	bool death;
	int deathtimer;
	bool doortext;
	int lockedtext;
	bool jumpability;
	bool ab_l, ab_r;

	std::list<Box> boxlist;
	std::list<Enemy> enemylist;


	void Init(Maps::Type *curmap, int mapx, int mapy, bool have_gun, bool right = 1)
	{
		map = curmap;
		x = mapx * 48 + 24;
		y = mapy * 48 + 25;
		l_or_r = right;
		xv = yv = 0;
		anim_state = 0;
		paused = 0;
		nopause = 0;
		hlight[0] = hlight[1] = hlight[2] = hlight[3] = 0;
		menu_choice = -1;
		nextmap = 0;
		gun = have_gun;
		gun_placed = curmap->gun_placed;
		gunpickuptext = 0;
		placedgunpos = {curmap->gunpos.x * 48 + 10, curmap->gunpos.y * 48 + 35};
		gunrot = 0;
		pr[0] = pr[1] = 0;
		prpos[0] = prpos[1] = {0,0};
		pra[0] = pra[1] = 0;
		boxlist = curmap->boxlist;
		death = 0;
		deathtimer = 0;
		doortext = 0;
		lockedtext = 0;
		jumpability = 1;
		ab_l = 1;
		ab_r = 1;
		enemylist = curmap->enemylist;

		Controls::UpdateNames();

		Shaders::s_t->Use();
		ScreenColor({0,0,0,0},{1,1,1,1});
	}

	void Tick()
	{
		static constexpr int vmax_h = 10, vmax_v = 10, walkspeed = 5, ladderspeed = 3, jumpforce = 20, proj_speed = 6;

		Audio::ListenerPos({float(x), float(y), -600});


		if (Events::KeyPressed(Controls::pause) && !nopause)
		{
			paused = !paused;
			Shaders::s_t->Use();
			if (paused)
			{
				ScreenColor({-0.25, -0.25, -0.25, 0}, {0.75, 0.75, 0.75, 1});
				hlight[0] = hlight[1] = hlight[2] = hlight[3] = 0;
			}
			else
			{
				ScreenColor({0,0,0,0},{1,1,1,1});
			}
		}
		if (paused)
		{
			int menu_pos = -1;
			for (int i = 0; i < 4; i++)
			{
				if (MouseInRect({menu_w / -2, 256 + menu_h * i}, {menu_w, menu_h}))
					menu_pos = i;
			}
			for (int i = 0; i < 4; i++)
			{
				if (i == menu_pos)
				{
					hlight[i] += 0.1;
					if (hlight[i] > 1)
						hlight[i] = 1;
				}
				else
				{
					hlight[i] -= 0.1;
					if (hlight[i] < 0)
						hlight[i] = 0;
				}
			}

			if (Events::MouseButtonPressed(0))
				menu_choice = menu_pos;
			return;
		}

		auto Kill = [&]() -> void
		{
			death = 1;
			nopause = 1;
			Sounds::death->Play();
		};

		auto SolidAtPoint = [](int xpos, int ypos) -> bool
		{
			return Maps::solid_table[map->AtPoint(xpos, ypos)];
		};

		auto SolidAtPointExcGrids = [](int xpos, int ypos) -> bool
		{
			int tmp = map->AtPoint(xpos, ypos);
			return Maps::solid_table[tmp] && (char(tmp) != 'X');
		};

		auto BoxAt = [](int xpos, int ypos, Box **p) -> bool
		{
			for (auto &it : boxlist)
			{
				switch (it.type)
				{
				  case 0:
					if ((xpos >= it.x - 22) &&
						(xpos <  it.x + 22) &&
						(ypos >= it.y - 22) &&
						(ypos <  it.y + 22))
					{
						*p = &it;
						return 1;
					}
					break;
				  case 1:
					if ((xpos >= it.x - 22) &&
						(xpos <  it.x + 22) &&
						(ypos >= it.y - 46) &&
						(ypos <  it.y + 46))
					{
						*p = &it;
						return 1;
					}
					break;
				  case 2:
					if ((xpos >= it.x - 46) &&
						(xpos <  it.x + 46) &&
						(ypos >= it.y - 46) &&
						(ypos <  it.y + 46))
					{
						*p = &it;
						return 1;
					}
					break;
				  case 3:
					if ((xpos >= it.x - 70) &&
						(xpos <  it.x + 70) &&
						(ypos >= it.y - 22) &&
						(ypos <  it.y + 22))
					{
						*p = &it;
						return 1;
					}
					break;
				}
			}
			return 0;
		};

		auto SolidAt = [=](int xoff, int yoff) -> bool
		{
			if (SolidAtPoint(x+xoff-9,  y+yoff-21) ||
				SolidAtPoint(x+xoff-9,  y+yoff+22) ||
				SolidAtPoint(x+xoff+10, y+yoff-21) ||
				SolidAtPoint(x+xoff+10, y+yoff+22))
				return 1;
			for (const auto &it : boxlist)
			{
				switch (it.type)
				{
				  case 0:
					if ((x+xoff > it.x - 31) &&
						(x+xoff < it.x + 29) &&
						(y+yoff > it.y - 45) &&
						(y+yoff < it.y + 41))
						return 1;
					break;
				  case 1:
					if ((x+xoff > it.x - 31) &&
						(x+xoff < it.x + 29) &&
						(y+yoff > it.y - 45-24) &&
						(y+yoff < it.y + 41+24))
						return 1;
					break;
				  case 2:
					if ((x+xoff > it.x - 31-24) &&
						(x+xoff < it.x + 29+24) &&
						(y+yoff > it.y - 45-24) &&
						(y+yoff < it.y + 41+24))
						return 1;
					break;
				  case 3:
					if ((x+xoff > it.x - 31-48) &&
						(x+xoff < it.x + 29+48) &&
						(y+yoff > it.y - 45) &&
						(y+yoff < it.y + 41))
						return 1;
					break;
				}
			}

			return 0;
		};

		auto SolidAtAbs = [=](int xoff, int yoff) -> bool
		{
			if (SolidAtPoint(xoff-9,  yoff-21) ||
				SolidAtPoint(xoff-9,  yoff+22) ||
				SolidAtPoint(xoff+10, yoff-21) ||
				SolidAtPoint(xoff+10, yoff+22))
				return 1;
			for (const auto &it : boxlist)
			{
				switch (it.type)
				{
				  case 0:
					if ((xoff > it.x - 31) &&
						(xoff < it.x + 29) &&
						(yoff > it.y - 45) &&
						(yoff < it.y + 41))
						return 1;
					break;
				  case 1:
					if ((xoff > it.x - 31) &&
						(xoff < it.x + 29) &&
						(yoff > it.y - 45-24) &&
						(yoff < it.y + 41+24))
						return 1;
					break;
				  case 2:
					if ((xoff > it.x - 31-24) &&
						(xoff < it.x + 29+24) &&
						(yoff > it.y - 45-24) &&
						(yoff < it.y + 41+24))
						return 1;
					break;
				  case 3:
					if ((xoff > it.x - 31-48) &&
						(xoff < it.x + 29+48) &&
						(yoff > it.y - 45) &&
						(yoff < it.y + 41))
						return 1;
					break;
				}
			}

			return 0;
		};

		auto BoxSolidAt = [=](int xoff, int yoff, Box *b) -> bool
		{
			switch (b->type)
			{
			  case 0:
				if (SolidAtPoint(b->x+xoff-22, b->y+yoff-22) ||
				    SolidAtPoint(b->x+xoff-22, b->y+yoff+21) ||
				    SolidAtPoint(b->x+xoff+21, b->y+yoff-22) ||
				    SolidAtPoint(b->x+xoff+21, b->y+yoff+21))
				return 1;
				break;
			  case 1:
				if (SolidAtPoint(b->x+xoff-22, b->y+yoff-46) ||
				    SolidAtPoint(b->x+xoff-22, b->y+yoff+45) ||
				    SolidAtPoint(b->x+xoff-22, b->y+yoff+ 0) ||
				    SolidAtPoint(b->x+xoff+21, b->y+yoff+ 0) ||
				    SolidAtPoint(b->x+xoff+21, b->y+yoff-46) ||
				    SolidAtPoint(b->x+xoff+21, b->y+yoff+45))
				return 1;
				break;
			  case 2:
				if (SolidAtPoint(b->x+xoff-46, b->y+yoff-46) ||
				    SolidAtPoint(b->x+xoff-46, b->y+yoff+45) ||
				    SolidAtPoint(b->x+xoff-46, b->y+yoff+ 0) ||
				    SolidAtPoint(b->x+xoff+45, b->y+yoff+ 0) ||
				    SolidAtPoint(b->x+xoff- 0, b->y+yoff-46) ||
				    SolidAtPoint(b->x+xoff+ 0, b->y+yoff+45) ||
				    SolidAtPoint(b->x+xoff+45, b->y+yoff-46) ||
				    SolidAtPoint(b->x+xoff+45, b->y+yoff+45))
				return 1;
				break;
			  case 3:
				if (SolidAtPoint(b->x+xoff-70, b->y+yoff-22) ||
				    SolidAtPoint(b->x+xoff-70, b->y+yoff+21) ||
				    SolidAtPoint(b->x+xoff-35, b->y+yoff-22) ||
				    SolidAtPoint(b->x+xoff-35, b->y+yoff+21) ||
				    SolidAtPoint(b->x+xoff+ 0, b->y+yoff-22) ||
				    SolidAtPoint(b->x+xoff+ 0, b->y+yoff+21) ||
				    SolidAtPoint(b->x+xoff+34, b->y+yoff-22) ||
				    SolidAtPoint(b->x+xoff+34, b->y+yoff+21) ||
				    SolidAtPoint(b->x+xoff+69, b->y+yoff-22) ||
				    SolidAtPoint(b->x+xoff+69, b->y+yoff+21))
				return 1;
				break;
			}

			static constexpr int xarr[4] {22, 22, 46, 70};
			static constexpr int yarr[4] {22, 46, 46, 22};

			for (auto &it : boxlist)
			{
				if (&it == b)
					continue;
				if (b->x+xoff+xarr[b->type] > it.x - xarr[it.type] &&
					b->y+yoff+yarr[b->type] > it.y - yarr[it.type] &&
					b->x+xoff-xarr[b->type] < it.x + xarr[it.type] &&
					b->y+yoff-yarr[b->type] < it.y + yarr[it.type])
				return 1;
			}

			return 0;
		};

		if (death)
		{
			Shaders::s_t->Use();
			ScreenColor({-0.25f * deathtimer / 256.0f, -0.25f * deathtimer / 256.0f, -0.25f * deathtimer / 256.0f, 0}, {1 - 0.25f * deathtimer / 256.0f, 1 - 0.45f * deathtimer / 256.0f, 1 - 0.45f * deathtimer / 256.0f, 1});

			deathtimer++;

			if (!SolidAt(0, 1))
				y++;

			if (deathtimer > 64)
			{
				if (deathtimer > 256)
					deathtimer = 256;

				if (Events::WhatKeyIsPressed(0))
					menu_choice = 1;
			}

			return;
		}


		int curblock = map->AtPoint(x, y+10);

		int hc = int(Events::KeyDown(Controls::right)) - int(Events::KeyDown(Controls::left));

		if (!ab_l && hc < 0) hc = 0;
		if (!ab_r && hc > 0) hc = 0;

		if (hc)
		{
			if (char(curblock) != 'L')
			{
				if ((hc > 0 && xv < walkspeed) || (hc < 0 && xv > -walkspeed))
					xv += hc;
			}
			else
			{
				if ((hc > 0 && xv < ladderspeed) || (hc < 0 && xv > -ladderspeed))
					xv += hc;
			}
		}
		else
		{
			if (xv)
			{
				if (xv > 0)
					xv--;
				else
					xv++;
			}
		}

		Box *curbox = 0;
		int boxox = 0, boxoy = 0;

		yv += 1;

		if (char(curblock) != 'L')
		{
			if (SolidAt(0, 1))
			{
				if (BoxAt(x, y+23, &curbox))
				{
					boxox = curbox->x;
					boxoy = curbox->y;
				}

				if (Events::KeyPressed(Controls::up) && jumpability)
				{
					yv = -jumpforce;
					Sounds::jump->Play();
				}
			}
		}

		switch (char(curblock))
		{
		  case 'd':
			{
				if (Events::KeyPressed(Controls::down))
					lockedtext = 120;
			}
			break;
		  case 'D':
			{
				if (std::abs(yv) <= 1 && SolidAt(0, 1))
				{
					doortext = 1;
					if (Events::KeyPressed(Controls::down))
					{
						nextmap = 1;
						Sounds::door->Play();
					}
				}
			}
			break;
		  case 'S':
			{
				Kill();
			}
			break;
		  case 'L':
			{
				yv -= 1;
				int vc = int(Events::KeyDown(Controls::down)) - int(Events::KeyDown(Controls::up));
				if (vc)
				{
					if ((vc > 0 && yv < ladderspeed) || (vc < 0 && yv > -ladderspeed))
						yv += vc;
				}
				else
				{
					if (yv)
					{
						if (yv > 0)
							yv--;
						else
							yv++;
					}
				}

				if (yv)
				{
					if (yv > ladderspeed)
						yv = ladderspeed;
					else if (yv < -ladderspeed)
						yv = -ladderspeed;
				}
			}
		}


		static constexpr float mag_force = 1;

		for (auto &it : boxlist)
		{
			for (auto &jt : boxlist)
			{
                if (it.charge == 0 || jt.charge == 0 || &it == &jt)
					continue;
				float force = mag_force * it.charge * jt.charge;
				float xx = it.x - jt.x;
				float yy = it.y - jt.y;
				float d = std::sqrt(xx*xx + yy*yy);
				if (d == 0)
					continue;
				it.xv += force * xx / d;
				it.yv += force * yy / d;
			}

		}

		for (auto &it : boxlist)
		{
			it.xv /= 1.1;
			it.yv /= 1.1;
			if (std::abs(it.xv) < 0.5f) it.xv = 0;
			if (std::abs(it.yv) < 0.5f) it.yv = 0;

			if (it.xv > 0)
			{
				int v;
				if (it.xv > vmax_h)
					v = it.xv = vmax_h;
				else
					v = it.xv;
				while (v-- > 0)
				{
					if (!BoxSolidAt(1, 0, &it))
						it.x += 1;
					else
					{
						it.xv = 0;
						break;
					}
				}
			}
			else if (it.xv < 0)
			{
				int v;
				if (it.xv < -vmax_h)
					v = it.xv = -vmax_h;
				else
					v = it.xv;

				while (v++ < 0)
				{
					if (!BoxSolidAt(-1, 0, &it))
						it.x -= 1;
					else
					{
						it.xv = 0;
						break;
					}
				}
			}

			if (it.yv > 0)
			{
				int v;
				if (it.yv > vmax_v)
					v = it.yv = vmax_v;
				else
					v = it.yv;

				while (v-- > 0)
				{
					if (!BoxSolidAt(0, 1, &it))
						it.y += 1;
					else
					{
						it.yv = 0;
						break;
					}
				}
			}
			else if (it.yv < 0)
			{
				int v;
				if (it.yv < -vmax_v)
					v = it.yv = -vmax_v;
				else
					v = it.yv;

				while (v++ < 0)
				{
					if (!BoxSolidAt(0, -1, &it))
						it.y -= 1;
					else
					{
						it.yv = 0;
						break;
					}
				}
			}

			it.yv++;

			if (it.xv > 0)
			{
				if (BoxSolidAt(1, 0, &it))
					it.xv = 0;
			}
			else
			{
				if (BoxSolidAt(-1, 0, &it))
					it.xv = 0;
			}
			if (it.yv > 0)
			{
				if (BoxSolidAt(0, 1, &it))
					it.yv = 0;
			}
			else
			{
				if (BoxSolidAt(0, -1, &it))
					it.yv = 0;
			}
		}


		if (curbox)
		{
			if (!SolidAt(curbox->x - boxox, curbox->y - boxoy))
			{
				x += curbox->x - boxox;
				y += curbox->y - boxoy;
			}
		}


		if (gun)
		{
			GFX::Vec<int, 2> center{GFX::WindowW_half(), GFX::WindowH_half() + pl_y};
			if (Events::MouseX() == center.x && Events::MouseY() == center.y)
				gunrot = fPi_half;
			else
				gunrot = std::atan2(Events::MouseY() - center.y, Events::MouseX() - center.x);

			if (Events::MouseButtonPressed(0) && !pr[0])
			{
                pr[0] = 1;
                prpos[0].x = x+1;
                prpos[0].y = y+8;
                pra[0] = gunrot;
                l_or_r = gunrot < fPi_half && gunrot >= -fPi_half;
                static unsigned int var = 0;
				if (var++ % 2 == 0)
					Sounds::shot1->Play();
				else
					Sounds::shot1a->Play();
			}
			if (Events::MouseButtonPressed(2) && !pr[1])
			{
                pr[1] = 1;
                prpos[1].x = x+1;
                prpos[1].y = y+8;
                pra[1] = gunrot;
                l_or_r = gunrot <= fPi_half && gunrot > -fPi_half;
                static unsigned int var = 0;
				if (var++ % 2 == 0)
					Sounds::shot2->Play();
				else
					Sounds::shot2a->Play();
			}

			static unsigned int wh_s_var = 0;
			for (int i = 0; i < 2; i++)
			{
				if (!pr[i])
					continue;
				prpos[i].x += std::cos(pra[i]) * proj_speed;
				prpos[i].y += std::sin(pra[i]) * proj_speed;

				if (SolidAtPointExcGrids(prpos[i].x, prpos[i].y))
				{
					pr[i] = 0;
					if (wh_s_var++ % 2 == 0)
					{
						Sounds::wallhit->Play();
						Sounds::wallhit->Pos({prpos[i].x, prpos[i].y, 0});
						Sounds::wallhit->ApplyChanges();
					}
					else
					{
						Sounds::wallhita->Play();
						Sounds::wallhita->Pos({prpos[i].x, prpos[i].y, 0});
						Sounds::wallhita->ApplyChanges();
					}
				}
				if ((prpos[i].x-x)*(prpos[i].x-x)+(prpos[i].y-y)*(prpos[i].y-y) > 4000000)
					pr[i] = 0;
			}


			Box *tmp;

			if (pr[0])
			{
				Sounds::shot1->Pos({prpos[0].x, prpos[0].y, 0});
				Sounds::shot1a->Pos({prpos[0].x, prpos[0].y, 0});
				Sounds::shot1->ApplyChanges();
				Sounds::shot1a->ApplyChanges();
				if (BoxAt(prpos[0].x, prpos[0].y, &tmp))
				{
					static unsigned int var = 0;
					if (var++ % 2 == 0)
					{
						Sounds::blhit1->Play();
						Sounds::blhit1->Pos({prpos[0].x, prpos[0].y, 0});
						Sounds::blhit1->ApplyChanges();
					}
					else
					{
						Sounds::blhit1a->Play();
						Sounds::blhit1a->Pos({prpos[0].x, prpos[0].y, 0});
						Sounds::blhit1a->ApplyChanges();
					}
					tmp->charge++;
					if (tmp->charge > 2)
						tmp->charge = 2;
					pr[0] = 0;
				}
			}
			if (pr[1])
			{
				Sounds::shot2->Pos({prpos[1].x, prpos[1].y, 0});
				Sounds::shot2a->Pos({prpos[1].x, prpos[1].y, 0});
				Sounds::shot2->ApplyChanges();
				Sounds::shot2a->ApplyChanges();
				if (BoxAt(prpos[1].x, prpos[1].y, &tmp))
				{
					static unsigned int var = 0;
					if (var++ % 2 == 0)
					{
						Sounds::blhit2->Play();
						Sounds::blhit2->Pos({prpos[0].x, prpos[0].y, 0});
						Sounds::blhit2->ApplyChanges();
					}
					else
					{
						Sounds::blhit2a->Play();
						Sounds::blhit2a->Pos({prpos[0].x, prpos[0].y, 0});
						Sounds::blhit2a->ApplyChanges();
					}
					tmp->charge--;
					if (tmp->charge < -2)
						tmp->charge = -2;
					pr[1] = 0;
				}
			}

		}
		else
		{
			if ((x-placedgunpos.x)*(x-placedgunpos.x)+(y-placedgunpos.y)*(y-placedgunpos.y) < 800)
			{
				if (Events::KeyPressed(Controls::use))
				{
					gun = 1;
					gun_placed = 0;
					Sounds::pickup->Play();
				}
				else
					gunpickuptext = 1;
			}
		}

		if (xv > 0)
		{
			l_or_r = 1;

			int v;
			if (xv > vmax_h)
				v = vmax_h;
			else
				v = xv;
			while (v-- > 0)
			{
				if (!SolidAt(1, 0))
					x += 1;
				else
					break;
			}
			if (SolidAt(1, 0))
				xv = 0;
		}
		else if (xv < 0)
		{
			l_or_r = 0;

			int v;
			if (xv < -vmax_h)
				v = -vmax_h;
			else
				v = xv;

			while (v++ < 0)
			{
				if (!SolidAt(-1, 0))
					x -= 1;
				else
					break;
			}
			if (SolidAt(-1, 0))
				xv = 0;
		}

		if (yv > 0)
		{
			int v;
			if (yv > vmax_v)
				v = vmax_v;
			else
				v = yv;

			while (v-- > 0)
			{
				if (!SolidAt(0, 1))
					y += 1;
				else
					break;
			}
			if (SolidAt(0, 1))
				yv = 0;
		}
		else if (yv < 0)
		{
			int v;
			if (yv < -vmax_v)
			{
				v = -vmax_v;
				if (!Events::KeyDown(Controls::up))
					yv = v;
			}
			else
				v = yv;

			while (v++ < 0)
			{
				if (!SolidAt(0, -1))
					y -= 1;
				else
					break;
			}
			if (SolidAt(0, -1))
				yv = 0;
		}

		if (SolidAt(0, 0))
		{
			bool ok = 0;
			for (int i = 0; i < 11; i++)
			{
				if (!SolidAt(0, -i))
				{
					y -= i;
					ok = 1;
					break;
				}
				else if (!SolidAt(-i, 0))
				{
					x -= i;
					ok = 1;
					break;
				}
				else if (!SolidAt(i, 0))
				{
					x += i;
					ok = 1;
					break;
				}
			}
			if (!ok)
			{
				Kill();
			}
		}

		if (yv)
		{
			if (yv < 0)
				anim_state = 3;
			else
				anim_state = 4;
		}
		else
		{
			if (xv)
			{
				static unsigned int anim_c = 0;
				anim_state = (anim_c++ % 30 < 15) + 1;
			}
			else
				anim_state = 0;
		}





		for (auto i = enemylist.begin(); i != enemylist.end();)
		{
			Enemy &it = *i;
			if (SolidAtAbs(it.x,it.y))
			{
				i = enemylist.erase(i);
				Sounds::kill->Play();
				continue;
			}

			if (it.l_or_r)
			{
				if (!SolidAtAbs(it.x+1, it.y) && SolidAtAbs(it.x+1, it.y+1))
					it.x++;
				else
					it.l_or_r = 0;
			}
			else
			{
				if (!SolidAtAbs(it.x-1, it.y) && SolidAtAbs(it.x-1, it.y+1))
					it.x--;
				else
					it.l_or_r = 1;
			}

			it.yv++;
			int v;
			if (it.yv > vmax_v)
				v = it.yv = vmax_v;
			else
				v = it.yv;

			if (it.yv > 0)
			{
				while (v--)
				{
					if (!SolidAtAbs(it.x, it.y+1))
						it.y++;
				}

				if (SolidAtAbs(it.x, it.y+1))
					it.yv = 0;
			}

			if ((it.x-x)*(it.x-x)+(it.y-y)*(it.y-y) < 256)
			{
				Kill();
				break;
			}
			i++;
		}
	}

	void Draw()
	{
		for (const auto &it : enemylist)
		{
			Draw::PushT({float(it.x-x - 9), float(it.y-y+pl_y - 19)}, {20, 42}, {float(227 + 16 * it.l_or_r), 75}, {10, 21}, -8);
		}

		for (const auto &it : boxlist)
		{
			switch (it.type)
			{
			  case 0:
				Draw::PushT({float(it.x-x - 24), float(it.y-y+pl_y - 24)}, {48, 48}, {float(216 + (it.charge != 0 ? (it.charge < 0 ? 0 : 96) + 96 : 0)), float(72 * (it.charge == 2 || it.charge == -2))}, {24, 24}, -9);
				break;
			  case 1:
				Draw::PushT({float(it.x-x - 24), float(it.y-y+pl_y - 48)}, {48, 96}, {float(240 + (it.charge != 0 ? (it.charge < 0 ? 0 : 96) + 96 : 0)), float(72 * (it.charge == 2 || it.charge == -2))}, {24, 48}, -9);
				break;
			  case 2:
				Draw::PushT({float(it.x-x - 48), float(it.y-y+pl_y - 48)}, {96, 96}, {float(264 + (it.charge != 0 ? (it.charge < 0 ? 0 : 96) + 96 : 0)), float(72 * (it.charge == 2 || it.charge == -2))}, {48, 48}, -9);
				break;
			  case 3:
				Draw::PushT({float(it.x-x - 72), float(it.y-y+pl_y - 24)}, {144,48}, {float(240 + (it.charge != 0 ? (it.charge < 0 ? 0 : 96) + 96 : 0)), 48 + float(72 * (it.charge == 2 || it.charge == -2))}, {72, 24}, -9);
				break;
			}
		}

		Draw::PushT({-9, pl_y-19}, {20, 42}, {float(195 + 16 * l_or_r), float(27 + 24 * anim_state)}, {10, 21}, -10);
		if (gun)
		{
			Draw::PushT({+1, pl_y+8}, {48, 32}, {float(192 + 24*l_or_r), 144}, {24, 16}, {float(32 - l_or_r*16), 16}, gunrot - fPi * !l_or_r, -11);
			for (int i = 0; i < 2; i++)
			{
				if (!pr[i])
					continue;
				Draw::PushT({prpos[i].x - x, prpos[i].y - y + pl_y}, {16, 16}, {float(224 + 8*i), 136}, {8, 8}, {8, 8}, pra[i], -12);
			}
		}
		if (gun_placed)
		{
			Draw::PushT({float(placedgunpos.x - x - 8), float(placedgunpos.y - y + pl_y - 3)}, {48, 16}, {192, 160}, {24, 8}, -11);
			if (gunpickuptext)
			{
				gunpickuptext = 0;
				Draw::Text({0,0}, Str({"Press ", Controls::names[4], " to pick up the gun"}), {1, 1, 1, 1}, 0, -12);
			}
		}
		if (doortext)
		{
			doortext = 0;
			Draw::Text({0,0}, Str({"Press ", Controls::names[1], " to enter"}), {1, 1, 1, 1}, 0, -12);
		}
		if (lockedtext)
		{
			Draw::Text({0,pl_y+256-32}, "\"It's locked!\"", {1, 1, 0.5, 1}, 0, -12);
			lockedtext--;
		}
		if (paused)
		{
			Draw::Text({0, -128}, "PAUSED", {1.0f/0.75f+0.25f, 1.0f/0.75f+0.25f, 0.5f/0.75f+0.25f, 1}, 0);
			Draw::Text({0, 256 + menu_h / 2 - 16 + menu_h * 0}, "Continue",      {1.0f/0.75f+0.25f, (1 - hlight[0] / 2)/0.75f+0.25f, (1 - hlight[0])/0.75f+0.25f, 1}, 0, -40);
			Draw::Text({0, 256 + menu_h / 2 - 16 + menu_h * 1}, "Restart room",  {1.0f/0.75f+0.25f, (1 - hlight[1] / 2)/0.75f+0.25f, (1 - hlight[1])/0.75f+0.25f, 1}, 0, -40);
			Draw::Text({0, 256 + menu_h / 2 - 16 + menu_h * 2}, "Back to menu",  {1.0f/0.75f+0.25f, (1 - hlight[2] / 2)/0.75f+0.25f, (1 - hlight[2])/0.75f+0.25f, 1}, 0, -40);
			Draw::Text({0, 256 + menu_h / 2 - 16 + menu_h * 3}, "Quit",          {1.0f/0.75f+0.25f, (1 - hlight[3] / 2)/0.75f+0.25f, (1 - hlight[3])/0.75f+0.25f, 1}, 0, -40);
		}
		if (deathtimer)
		{
			Draw::Text({0, -64}, "You're dead", {1, 1, 1, deathtimer / 256.0f}, 0, -40.5);
		}
		if (paused || gun)
			Draw::Mouse();
	}
}

namespace Menu
{
	Main::StateFunc Tick()
	{
		Sounds::maintheme->Stop();
		constexpr int menulen = 3;

		float hlight[menulen]{};

		while (1)
		{
			Events::PreTick();

			int menu_pos = -1;
			for (int i = 0; i < menulen; i++)
			{
				if (MouseInRect({menu_w / -2, menu_y + menu_h * i}, {menu_w, menu_h}))
					menu_pos = i;
			}
			for (int i = 0; i < menulen; i++)
			{
				if (i == menu_pos)
				{
					hlight[i] += 0.1;
					if (hlight[i] > 1)
						hlight[i] = 1;
				}
				else
				{
					hlight[i] -= 0.1;
					if (hlight[i] < 0)
						hlight[i] = 0;
				}
			}

			//////////////////
			//////////////////
			//////////////////


			Shaders::s_t->Use();
			camc.SetUniFinal(UniLoc::t_mf);
			rq_t->Clear();


			Draw::Text({0, menu_y + menu_h / 2 - 16 + menu_h * 0}, "New game",   {1, 1 - hlight[0] / 2, 1 - hlight[0], 1}, 0);
			Draw::Text({0, menu_y + menu_h / 2 - 16 + menu_h * 1}, "Controls",   {1, 1 - hlight[1] / 2, 1 - hlight[1], 1}, 0);
			Draw::Text({0, menu_y + menu_h / 2 - 16 + menu_h * 2}, "Quit",       {1, 1 - hlight[2] / 2, 1 - hlight[2], 1}, 0);
			Draw::Text({0, GFX::WindowH_half()-32}, "A game by HolyBlackCat (std-cin@mail.ru), Apr 20  2015", {0.25, 0.25, 0.25, 1}, 0);

			Draw::PushT({-256, -128}, {512, 128}, {288, 232}, {256, 64}, {0,0,0,0}, {1,1,1,1}, -1);

			Draw::Mouse();

			rq_t->Send();
			rq_t->DrawQuads();
			Shaders::s_c->Use();
			camc.SetUniFinal(UniLoc::c_mf);
			rq_c->Clear();



			rq_c->Send();
			rq_c->DrawQuads();
			Events::PostTick();

			if (Events::MouseButtonPressed(0) && menu_pos != -1)
			{
				switch (menu_pos)
				{
				  case 0:
					return Prog::States::World;
					break;
				  case 1:
					return Prog::States::Config;
					break;
				  case 2:
					Main::Exit();
					break;
				}
			}

		}
	}
}

namespace Config
{
	void SaveControls()
	{
		std::ofstream out_f("controlsconfig");
		if (out_f)
		{
			out_f << Controls::up      << '\n';
			out_f << Controls::down    << '\n';
			out_f << Controls::left    << '\n';
			out_f << Controls::right   << '\n';
			out_f << Controls::use     << '\n';
			out_f << Controls::pause;
			out_f.close();
		}
	}

	Main::StateFunc Tick()
	{
		float backhlight = 0;
		float hlight[Controls::keys]{};
		int active = -1;

		Controls::UpdateNames();



		while (1)
		{
			Events::PreTick();

			static constexpr int menu_w = 256, menu_h = 48, menu_y = -48*3, back_y = 192-32;
			int menu_pos = -1;
			bool back_hovered = 0;
			if (active == -1)
			{
				for (int i = 0; i < Controls::keys; i++)
				{
					if (MouseInRect({menu_w / -2, menu_y + menu_h * i}, {menu_w, menu_h}))
						menu_pos = i;
				}
				if (MouseInRect({menu_w / -2, back_y - menu_h / 2 + 8}, {menu_w, menu_h}))
					back_hovered = 1;
			}
			else
			{
				unsigned int k;
				if (Events::WhatKeyIsPressed(&k))
				{
					if (k != Controls::up && k != Controls::down && k != Controls::left && k != Controls::right && k != Controls::use && k != Controls::pause)
					{
						switch (active)
						{
						  case 0:
							Controls::up = k;
							Controls::UpdateNames();
							break;
						  case 1:
							Controls::down = k;
							Controls::UpdateNames();
							break;
						  case 2:
							Controls::left = k;
							Controls::UpdateNames();
							break;
						  case 3:
						  	Controls::right = k;
							Controls::UpdateNames();
							break;
						  case 4:
							Controls::use = k;
							Controls::UpdateNames();
							break;
						  case 5:
							Controls::pause = k;
							Controls::UpdateNames();
							break;
						}
						SaveControls();
					}
					active = -1;
				}
			}

			if (Events::MouseButtonPressed(0) && menu_pos != -1)
				active = menu_pos;
			if (Events::MouseButtonPressed(0) && back_hovered)
			{
				Events::PostTick();
				return Prog::States::Menu;
			}

			for (int i = 0; i < Controls::keys; i++)
			{
				if (i == menu_pos)
				{
					hlight[i] += 0.1;
					if (hlight[i] > 1)
						hlight[i] = 1;
				}
				else
				{
					hlight[i] -= 0.1;
					if (hlight[i] < 0)
						hlight[i] = 0;
				}
			}
			if (back_hovered)
			{
				backhlight += 0.1;
				if (backhlight > 1)
					backhlight = 1;
			}
			else
			{
				backhlight -= 0.1;
				if (backhlight < 1)
					backhlight = 0;
			}


			Shaders::s_t->Use();
			camc.SetUniFinal(UniLoc::t_mf);
			rq_t->Clear();

			Draw::Text({0, -192}, "CONTROLS", {1,1,0.5,1}, 0);
			Draw::Text({0, menu_y + menu_h * 0 + menu_h / 2 - 16}, "Up: ",    {1, 1 - hlight[0] / 2, 1 - hlight[0], 1}, 1);
			Draw::Text({0, menu_y + menu_h * 0 + menu_h / 2 - 16}, Controls::names[0], active == 0 ? GFX::Vec<float, 4>{0,0.75,1,1} : GFX::Vec<float, 4>{0.75,0.75,0.75,1}, -1);
			Draw::Text({0, menu_y + menu_h * 1 + menu_h / 2 - 16}, "Down: ",  {1, 1 - hlight[1] / 2, 1 - hlight[1], 1}, 1);
			Draw::Text({0, menu_y + menu_h * 1 + menu_h / 2 - 16}, Controls::names[1], active == 1 ? GFX::Vec<float, 4>{0,0.75,1,1} : GFX::Vec<float, 4>{0.75,0.75,0.75,1}, -1);
			Draw::Text({0, menu_y + menu_h * 2 + menu_h / 2 - 16}, "Left: ",  {1, 1 - hlight[2] / 2, 1 - hlight[2], 1}, 1);
			Draw::Text({0, menu_y + menu_h * 2 + menu_h / 2 - 16}, Controls::names[2], active == 2 ? GFX::Vec<float, 4>{0,0.75,1,1} : GFX::Vec<float, 4>{0.75,0.75,0.75,1}, -1);
			Draw::Text({0, menu_y + menu_h * 3 + menu_h / 2 - 16}, "Right: ", {1, 1 - hlight[3] / 2, 1 - hlight[3], 1}, 1);
			Draw::Text({0, menu_y + menu_h * 3 + menu_h / 2 - 16}, Controls::names[3], active == 3 ? GFX::Vec<float, 4>{0,0.75,1,1} : GFX::Vec<float, 4>{0.75,0.75,0.75,1}, -1);
			Draw::Text({0, menu_y + menu_h * 4 + menu_h / 2 - 16}, "Use: ",   {1, 1 - hlight[4] / 2, 1 - hlight[4], 1}, 1);
			Draw::Text({0, menu_y + menu_h * 4 + menu_h / 2 - 16}, Controls::names[4], active == 4 ? GFX::Vec<float, 4>{0,0.75,1,1} : GFX::Vec<float, 4>{0.75,0.75,0.75,1}, -1);
			Draw::Text({0, menu_y + menu_h * 5 + menu_h / 2 - 16}, "Pause: ", {1, 1 - hlight[5] / 2, 1 - hlight[5], 1}, 1);
			Draw::Text({0, menu_y + menu_h * 5 + menu_h / 2 - 16}, Controls::names[5], active == 5 ? GFX::Vec<float, 4>{0,0.75,1,1} : GFX::Vec<float, 4>{0.75,0.75,0.75,1}, -1);
			Draw::Text({0, back_y}, "Back", {1, 1 - backhlight / 2, 1 - backhlight, 1}, 0);
			Draw::Mouse();

			rq_t->Send();
			rq_t->DrawQuads();
			Shaders::s_c->Use();
			camc.SetUniFinal(UniLoc::c_mf);
			rq_c->Clear();



			rq_c->Send();
			rq_c->DrawQuads();
			Events::PostTick();
		}
	}
}

namespace World
{
	Main::StateFunc Tick(int stage = -1)
	{

		while (1)
		{
		  restart:
			switch (stage)
			{
			  case -1:
				{

					Utils::TimerStatic timer;
					while (timer() < 3)
					{
						Events::PreTick();
						Events::PostTick();
					}
					timer.~TimerStatic(); new(&timer) Utils::TimerStatic;
					float t;
					Sounds::startbeep->Play();
					/*
					while ((t = timer()) < 8)
					{
						Events::PreTick();
						Events::PostTick();
					}
					timer.~TimerStatic(); new(&timer) Utils::TimerStatic;*/
					Sounds::startbeep->Play();
					bool s_flag = 0;
					while ((t = timer()) < 9)
					{
						Events::PreTick();

						Shaders::s_t->Use();
						camc.SetUniFinal(UniLoc::t_mf);
						rq_t->Clear();

						if (t < 1)
						{
							ScreenColor({-1 + t*3, -1.5f + t*3, -2 + t*3, 0}, {1, 1, 1, 1});
						}
						else
						{
							if (!s_flag && t > 8)
							{
								s_flag = 1;
								Sounds::startbeep->Play();
							}
							t -= 7.5;
							ScreenColor({1 - t*3, 2 - t*3, 3 - t*3, 0}, {1, 1, 1, 1});
						}

						Draw::Text({-504, -64},
						"\" If the two charges have the same sign,\n"
						"    the electrostatic force between them\n"
						"                           is repulsive;\n"
						"\n"
						"  if they have different sign,\n"
						"    the electrostatic force between them\n"
						"                          is attractive. \"\n"
						"\n"
						"                       - The Coulomb's Law\n",
						{1,1,1,1}, -1);

						rq_t->Send();
						rq_t->DrawQuads();
						Shaders::s_c->Use();
						camc.SetUniFinal(UniLoc::c_mf);
						rq_c->Clear();



						rq_c->Send();
						rq_c->DrawQuads();
						Events::PostTick();
					}
					timer.~TimerStatic(); new(&timer) Utils::TimerStatic;

					while ((t = timer()) < 3)
					{
						Events::PreTick();
						Events::PostTick();
					}

					Shaders::s_t->Use();
					ScreenColor({0,0,0,0}, {1,1,1,1});

					stage++;
				}
				break;

			  default:
				{
					Maps::Type *curmap = Maps::rooms + stage;
					Player::Init(curmap, curmap->spawn.x, curmap->spawn.y, curmap->gun);

					int t = 0;

					bool ending = stage == (sizeof Maps::rooms) / (sizeof *Maps::rooms) - 1;
					if (!ending)
						Sounds::maintheme->Play();
					else
						Sounds::maintheme->Stop();

					while (1)
					{
						Events::PreTick();

						Player::Tick();


						Shaders::s_t->Use();
						camc.SetUniFinal(UniLoc::t_mf);
						rq_t->Clear();

						Draw::Map(Player::x, Player::y-pl_y, *curmap, !ending ? 'W' : '-');

						Player::Draw();


						if (stage == 0)
						{
							if (t == 256)
							{
								if (!Player::death && !Player::paused)
									ScreenColor({0,0,0,0}, {1,1,1,1});
								Player::nopause = 0;
								t += 1;
							}
							else if (t < 256)
							{
								if (!Player::death && !Player::paused)
									ScreenColor({0,0,0,0}, {t/256.0f,t/256.0f,t/256.0f,1});
								t += 1;
							}

							Draw::Text({0,-300}, Str({"Press ", Controls::names[2], " or ", Controls::names[3], " to move\n"
													  "Press ", Controls::names[0], " to jump\n"
													  "Use ", Controls::names[5], " to pause game"}), {1, 1, 1, 1}, 0, -12);
						}
						else if (stage == 1)
						{
							if (Player::gun)
								Draw::Text({0,-300}, "Press LMB to fire positive projectile\n"
													 "Press RMB to fire negative projectile\n"
													 "Shoot at bricks to change their charge", {1, 1, 1, 1}, 0, -12);
							else
								Draw::Text({0,-300}, Str({"Press ", Controls::names[0], " or ", Controls::names[1], " to climb stairs"}), {1, 1, 1, 1}, 0, -12);
						}
						else if (stage == 2)
						{
							Draw::Text({0,-300}, "You can shoot through grids", {1, 1, 1, 1}, 0, -12);
						}

						if (ending)
						{
                            Player::jumpability = Player::x < 49*48;
                            Player::ab_l = Player::x > 22*48 && Player::x < 51*48;
                            Player::ab_r = Player::x < 57*48;

                            if (Player::gun && Player::x > 52*48)
							{
								Player::gun = 0;
								Player::gun_placed = 1;
								Player::placedgunpos = GFX::Vec<int, 2>{51*48+16, 5*48+35};
								Sounds::pickup->Play();
							}

							if (!Player::ab_r)
							{
								if (t < 1024)
								{
									ScreenColor({t/1024.0f,t/1024.0f,t/1024.0f,0}, {1+t/1024.0f,1+t/1024.0f,1+t/512.0f,1});
									t++;
								}
								else
								{
									stage++;
									goto restart;
								}
							}
						}


						rq_t->Send();
						rq_t->DrawQuads();
						Shaders::s_c->Use();
						camc.SetUniFinal(UniLoc::c_mf);
						rq_c->Clear();

						rq_c->Send();
						rq_c->DrawQuads();
						Events::PostTick();

						if (Player::nextmap == 1)
						{
							Player::nextmap = 0;
							stage++;
							goto restart;
						}

						if (Player::menu_choice != -1)
						{
                            switch (Player::menu_choice)
                            {
			  				  case 0:
			  				  	Shaders::s_t->Use();
								ScreenColor({0,0,0,0},{1,1,1,1});
								Player::paused = 0;
								Player::menu_choice = -1;
								break;
							  case 1:
								goto restart;
								break;
							  case 2:
							  	Shaders::s_t->Use();
								ScreenColor({0,0,0,0},{1,1,1,1});
								return Prog::States::Menu();
								break;
							  case 3:
								Main::Exit();
								break;
                            }
						}
					}
				}
				break;

			  case (sizeof Maps::rooms) / (sizeof *Maps::rooms):
				{
					ScreenColor({0,0,0,0},{1,1,1,1});

					int t = 0;

					Sounds::victory->Play();

					while (1)
					{
						Events::PreTick();

						Shaders::s_c->Use();
						camc.SetUniFinal(UniLoc::c_mf);
						rq_c->Clear();

						Draw::PushC({float(-GFX::WindowW_half()), float(-GFX::WindowH_half())}, {float(GFX::WindowW_m_of_2()), float(GFX::WindowH_m_of_2())}, {1,1,1,1}, 0);

						rq_c->Send();
						rq_c->DrawQuads();
						Shaders::s_t->Use();
						camc.SetUniFinal(UniLoc::t_mf);

						rq_t->Clear();

						if (t < 512)
						{
							t++;
						}
						else if (Events::WhatKeyIsPressed(0))
							return Prog::States::Menu;

						Draw::PushT({-256, -64}, {512, 128}, {192, 168}, {256, 64}, {0,0,0,0}, {1,1,1,t/512.0f}, -1);

						rq_t->Send();
						rq_t->DrawQuads();
						Events::PostTick();
					}
				}
				break;
			}
		}
		return Prog::States::Menu;
	}
}

namespace Prog
{
	namespace Tex
	{
		GFX::Texture *main;
	}

	void Init(int w, int h)
	{
		EXEC_ONCE;
		(void)w; (void)h;

		SDL_SetWindowMinimumSize(GFX::WinHandle(), 1024, 768);

		Events::MouseMode::Hide();

		GFX::Viewport_Fullscreen_m_of_2();

		Shaders::s_c = new GFX::Shader(ShaderCode::Colored::v, ShaderCode::Colored::f);
		Shaders::s_t = new GFX::Shader(ShaderCode::Textured::v, ShaderCode::Textured::f);
		GFX::Shader::LoadShaderList();
		Shaders::s_t->Use();

		GFX::Culling(0);
		GFX::Depth(1);

		GFX::ClearColor(0, 0, 0, 255);

		UniLoc::c_mf = glGetUniformLocation(*Shaders::s_c, "u_mat_final");    if (UniLoc::c_mf == -1) Main::Exit("Shader #0: Uniform `u_mat_final` not found.");
		UniLoc::t_mf = glGetUniformLocation(*Shaders::s_t, "u_mat_final");    if (UniLoc::t_mf == -1) Main::Exit("Shader #1: Uniform `u_mat_final` not found.");
		UniLoc::t_t  = glGetUniformLocation(*Shaders::s_t, "t");              if (UniLoc::t_t  == -1) Main::Exit("Shader #1: Uniform `t` not found.");
		UniLoc::t_co = glGetUniformLocation(*Shaders::s_t, "u_co");           if (UniLoc::t_co == -1) Main::Exit("Shader #1: Uniform `u_co` not found.");
		UniLoc::t_cm = glGetUniformLocation(*Shaders::s_t, "u_cm");           if (UniLoc::t_cm == -1) Main::Exit("Shader #1: Uniform `u_cm` not found.");

		GFX::ImageData img_tex("tex.tga");
		Tex::main = new GFX::Texture();
		GFX::Texture::SetActiveSlot(0);
		Tex::main->BindToActiveSlot();
		GFX::Texture::ActiveSlotData(img_tex);
		GFX::Texture::ActiveSlotParams(GFX::Interpolation::nearest, GFX::Wrap::transp);
		GFX::Texture::SetSlotAsUniform(UniLoc::t_t, 0);

		Audio::ListenerRot({0, 0, 1}, {0, -1, 0});
		Audio::ListenerVelocity({0,0,0});

		Sounds::Buffers::startbeep = new Audio::Buffer("audio\\startbeep.wav", 1);
		Sounds::startbeep = new Audio::Source(*Sounds::Buffers::startbeep, 0);
		Sounds::Buffers::door = new Audio::Buffer("audio\\door.wav", 1);
		Sounds::door = new Audio::Source(*Sounds::Buffers::door, 0);
		Sounds::Buffers::shot = new Audio::Buffer("audio\\shot.wav", 0);
		Sounds::shot1 = new Audio::Source(*Sounds::Buffers::shot, 0);
		Sounds::shot1a = new Audio::Source(*Sounds::Buffers::shot, 0);
		Sounds::shot2 = new Audio::Source(*Sounds::Buffers::shot, 0, 1, 0.5);
		Sounds::shot2a = new Audio::Source(*Sounds::Buffers::shot, 0, 1, 0.5);
		Sounds::Buffers::blhit = new Audio::Buffer("audio\\blhit.wav", 0);
		Sounds::blhit1 = new Audio::Source(*Sounds::Buffers::blhit, 0, 1, 1);
		Sounds::blhit1a = new Audio::Source(*Sounds::Buffers::blhit, 0, 1, 1);
		Sounds::blhit2 = new Audio::Source(*Sounds::Buffers::blhit, 0, 2, 0.5);
		Sounds::blhit2a = new Audio::Source(*Sounds::Buffers::blhit, 0, 2, 0.5);
		Sounds::Buffers::wallhit = new Audio::Buffer("audio\\wallhit.wav", 0);
		Sounds::wallhit = new Audio::Source(*Sounds::Buffers::wallhit, 0);
		Sounds::wallhita = new Audio::Source(*Sounds::Buffers::wallhit, 0);
		Sounds::Buffers::jump = new Audio::Buffer("audio\\jump.wav", 0);
		Sounds::jump = new Audio::Source(*Sounds::Buffers::jump, 0, 1, 0.5);
		Sounds::Buffers::kill = new Audio::Buffer("audio\\kill.wav", 1);
		Sounds::kill = new Audio::Source(*Sounds::Buffers::kill, 0, 1);
		Sounds::Buffers::death = new Audio::Buffer("audio\\death.wav", 1);
		Sounds::death = new Audio::Source(*Sounds::Buffers::death, 0);
		Sounds::Buffers::pickup = new Audio::Buffer("audio\\pickup.wav", 1);
		Sounds::pickup = new Audio::Source(*Sounds::Buffers::pickup, 0);
		Sounds::Buffers::victory = new Audio::Buffer("audio\\victory.wav", 1);
		Sounds::victory = new Audio::Source(*Sounds::Buffers::victory, 0, 0.25);
		Sounds::Buffers::maintheme = new Audio::Buffer("audio\\maintheme.wav", 1);
		Sounds::maintheme = new Audio::Source(*Sounds::Buffers::maintheme, 1, 0.25);
		alListenerf(AL_GAIN, 4);

		rq_c = new GFX::RenderingQueue<Layouts::pos3f_color4f>(0x10000);
		rq_t = new GFX::RenderingQueue<Layouts::pos3f_tex2f_co4f_cm4f>(0x10000);

		camc.ModelPositionRotation({0,0,0});
		camc.CameraPositionRotation({0,0,0}, {});
		camc.ViewParams_Flat(-64, 64);
		camc.ComputeFinalWithoutPrecalc();
		camc.ApplyRatioToMatrixes(GFX::WindowW_m_of_2(), GFX::WindowH_m_of_2());

		Maps::id_table   [(unsigned char) ' '] = 1;
		Maps::merge_table[(unsigned char) ' '] = 1;

		Maps::id_table   [(unsigned char) 'W'] = 2;
		Maps::merge_table[(unsigned char) 'W'] = 2;
		Maps::solid_table[(unsigned char) 'W'] = 1;

		Maps::id_table   [(unsigned char) 'D'] = 3;
		Maps::merge_table[(unsigned char) 'D'] = 3;

		Maps::id_table   [(unsigned char) 'B'] = 4;
		Maps::merge_table[(unsigned char) 'B'] = 4;
		Maps::solid_table[(unsigned char) 'B'] = 1;

		Maps::id_table   [(unsigned char) 'C'] = 5;
		Maps::merge_table[(unsigned char) 'C'] = 5;
		Maps::solid_table[(unsigned char) 'C'] = 1;

		Maps::id_table   [(unsigned char) 'b'] = 6;
		Maps::merge_table[(unsigned char) 'b'] = 6;

		Maps::id_table   [(unsigned char) 'd'] = 7;
		Maps::merge_table[(unsigned char) 'd'] = 7;

		Maps::id_table   [(unsigned char) 'L'] = 8;
		Maps::merge_table[(unsigned char) 'L'] = 8;

		Maps::id_table   [(unsigned char) 'S'] = 9;
		Maps::merge_table[(unsigned char) 'S'] = 9;

		Maps::id_table   [(unsigned char) 'X'] = 10;
		Maps::merge_table[(unsigned char) 'X'] = 10;
		Maps::solid_table[(unsigned char) 'X'] = 1;

		Maps::id_table   [(unsigned char) 'G'] = 11;
		Maps::merge_table[(unsigned char) 'G'] = 11;
		Maps::solid_table[(unsigned char) 'G'] = 1;

		Maps::id_table   [(unsigned char) '-'] = 12;
		Maps::merge_table[(unsigned char) '-'] = 12;


		std::ifstream inp_f("controlsconfig");

		if (!inp_f)
		{
			Config::SaveControls();
		}
		else
		{
			inp_f >> Controls::up;
			inp_f >> Controls::down;
			inp_f >> Controls::left;
			inp_f >> Controls::right;
			inp_f >> Controls::use;
			inp_f >> Controls::pause;

			if (!inp_f)
			{
				inp_f.close();
				Config::SaveControls();
			}
			else
				inp_f.close();
		}
	}

	void Resize(int w, int h)
	{
		(void)w; (void)h;
		GFX::Viewport_Fullscreen_m_of_2();
		camc.ApplyRatioToMatrixes(GFX::WindowW_m_of_2(), GFX::WindowH_m_of_2());
	}

	namespace States
	{
		Main::StateFunc Boot()
		{
			return Menu;
		}

		Main::StateFunc Menu()
		{
			return Menu::Tick();
		}

		Main::StateFunc Config()
		{
			return Config::Tick();
		}

		Main::StateFunc World()
		{
			return World::Tick();
		}
	}
}