#ifndef UTILS_H_INCL
#define UTILS_H_INCL

#include <cmath>
#include <cstring>
#include <chrono>
#include <random>
#include <sstream>
#include <type_traits>
#include "main.h"

#define EXEC_ONCE \
do {static bool EXEC_ONCE_FLAG = 0; if (EXEC_ONCE_FLAG) Main::Exit(Str({"Error: ", __func__, "() called twice."})); EXEC_ONCE_FLAG = 1;} while (0)

namespace Utils
{
	inline constexpr unsigned int DivCeil(unsigned int a, unsigned int b)
	{
		return (a + (b - 1)) / b ;
	}

	inline namespace TmplDefs
	{
		template <typename Type, unsigned int BlockLen = (4096 - sizeof (void *)) / sizeof (Type)>
		class BlockLinkedList
		{
			static_assert(std::is_pod<Type>::value, "BlockLinkedList is only for POD types.");
			struct Block
			{
				Type arr[BlockLen];
				Block *next;
			};

			unsigned int size;
			Block *first, **last;
		public:
			static constexpr unsigned int block_len = BlockLen;
			BlockLinkedList()
			  : size(0), first(0), last(&first) {}

			BlockLinkedList(const BlockLinkedList &o) : size(0), last(&first)
			{
				if (o.size == 0)
				{
					first = 0;
					last = &first;
				}

				const Block *o_p = o.first;

				while (1)
				{
					*last = new Block;
					if (o.size - size > BlockLen)
					{
						std::memcpy((*last)->arr, o_p, sizeof Block::arr);
						last = &((*last)->next);
						o_p = o_p->next;
						size += BlockLen;
					}
					else
					{
						std::memcpy((*last)->arr, o_p, (o.size - size) * sizeof (Type));
						(*last)->next = 0;
						size = o.size;
						return;
					}
				}
			}

			BlockLinkedList(BlockLinkedList &&o)
			{
				size = o.size;
				o.size = 0;
				first = o.first;
				o.first = 0;
				last = o.last;
				o.last = &o.first;
			}

			BlockLinkedList &operator=(const BlockLinkedList &o)
            {
            	this->~BlockLinkedList();
            	new(this) BlockLinkedList(o);
            	return *this;
            }

            BlockLinkedList &operator=(BlockLinkedList &&o)
            {
            	this->~BlockLinkedList();
            	new(this) BlockLinkedList((BlockLinkedList &&)o);
            	return *this;
            }

			~BlockLinkedList()
			{
				Block *cur = first, *tmp;
				while (cur)
				{
					tmp = cur;
					cur = cur->next;
					delete tmp;
				}
			}

			Type &operator[](unsigned int pos)
			{
				Block *p = first;
				int block_c = pos / BlockLen;
				while (block_c--)
				{
					p = p->next;
				}
				return p->arr[pos % BlockLen];
			}

			Type &operator[](unsigned int pos) const
			{
				const Block *p = first;
				int block_c = pos / BlockLen;
				while (block_c--)
				{
					p = p->next;
				}
				return p->arr[pos % BlockLen];
			}

			Type *PtrAt(unsigned int pos)
			{
				Block *p = first;
				int block_c = pos / BlockLen;
				while (block_c--)
				{
					p = p->next;
				}
				return p->arr + pos % BlockLen;
			}

			const Type *PtrAt(unsigned int pos) const
			{
				const Block *p = first;
				int block_c = pos / BlockLen;
				while (block_c--)
				{
					p = p->next;
				}
				return p->arr + pos % BlockLen;
			}

			Type &Add()
			{
				if (size + 1 < size)
					Main::Exit("BlockLinkedList length counter overflow.");

				if (DivCeil(size + 1, BlockLen) - DivCeil(size, BlockLen))
				{
					*last = new Block;
					last = &((*last)->next);
					*last = 0;
				}

				return operator[](size++);
			}

			void Remove(unsigned int n)
			{
				if (size - 1 > size)
					Main::Exit("BlockLinkedList length counter underflow.");
				std::memcpy(PtrAt(n), PtrAt(--size), sizeof (Type));
				if (DivCeil(size, BlockLen) == DivCeil(size + 1, BlockLen))
					return;
				unsigned int skip_c = size / BlockLen;
				last = &first;
				while (skip_c--)
					last = &((*last)->next);
				delete *last;
				*last = 0;
			}

			unsigned int Size() const
			{
				return size;
			}
		};
	}

	inline double Now()
	{
		return double(std::chrono::high_resolution_clock::now().time_since_epoch().count() * std::chrono::high_resolution_clock::period::num) / double(std::chrono::high_resolution_clock::period::den);
	}

	class Timer
	{
		using Clock = std::chrono::high_resolution_clock;
		Clock::duration t;
		bool of;
	public:
		Timer() : t(Clock::now() - Clock::time_point::min()), of(0) {}
		double operator()() // returns amount of seconds that passed since contruction or previous call
		{
			Clock::duration point = Clock::now() - Clock::time_point::min();

			if (point < t)
			{
				t = point;
				of = 1;
				return 0.0f;
			}

			double ret = double((point - t).count() * Clock::period::num) / double(Clock::period::den);
			t = point;
			return ret;
		}

		bool Overflow()
		{
			return of;
		}
	};

	class TimerStatic
	{
		using Clock = std::chrono::high_resolution_clock;
		Clock::duration t;
		bool of;
	public:
		TimerStatic() : t(Clock::now() - Clock::time_point::min()), of(0) {}
		double operator()() // returns amount of seconds that passed since contruction
		{
			Clock::duration point = Clock::now() - Clock::time_point::min();

			if (point < t)
			{
				t = point;
				of = 1;
				return 0.0f;
			}

			return double((point - t).count() * Clock::period::num) / double(Clock::period::den);
		}

		bool Overflow()
		{
			return of;
		}
	};

	inline void Wait(double secs)
	{
		TimerStatic t;
		volatile unsigned int v = 0;
		while (t() < secs && !t.Overflow())
			v++;
	}

	class UC
	{
	public:
		static std::stringstream ss;
		std::string str;
		template <class T> UC(T param) : str((ss.str(""), ss << param, ss.str())) {}
	};

	const char *Str(std::initializer_list<UC> data);

	namespace RNG
	{
		std::mt19937::result_type x32();
		std::mt19937_64::result_type x64();
		void x32_discard(unsigned int count);
		void x64_discard(unsigned int count);
		void x32_seed(std::mt19937::result_type p);
		void x64_seed(std::mt19937_64::result_type p);
	}

	static constexpr bool is_char_signed = char(-1) < 0;

	template <int X, int Y, typename T = float> inline T ScalarFromMat4Product(const T (*a)[4], const T (*b)[4]) // When X == 0 only b[0][...] is used
	{
		static_assert(X >= 0 && Y >= 0 && X < 4 && Y < 4, "Index is out of range.");
		return a[0][Y]*b[X][0] + a[1][Y]*b[X][1] + a[2][Y]*b[X][2] + a[3][Y]*b[X][3];
	}
}

extern const float fPi;
extern const double dPi;
extern const float fPi2;
extern const double dPi2;
extern const float fPi_half;
extern const double dPi_half;
extern const float fPi_quarter;
extern const double dPi_quarter;

inline float deg2rad(float x) {return x * fPi / 180.0f;}
inline float rad2deg(float x) {return x * 180.0f / fPi;}
inline double deg2rad(double x) {return x * dPi / 180.0;}
inline double rad2deg(double x) {return x * 180.0 / dPi;}

using Utils::Str;
using Utils::DivCeil;

#endif